@@ -8,6 +8,9 @@ from enum import Enum
 import uuid
 from fuzzy_json import loads as fuzzy_json_loads
 import logging
+from mcp.server.fastmcp.tools import Tool as MCPToolClass
+
+from notebook_intelligence.config import NBIConfig
 
 log = logging.getLogger(__name__)
 
@@ -30,6 +33,7 @@ class ResponseStreamDataType(str, Enum):
     LLMRaw = 'llm-raw'
     Markdown = 'markdown'
     MarkdownPart = 'markdown-part'
+    Image = 'image'
     HTMLFrame = 'html-frame'
     Button = 'button'
     Anchor = 'anchor'
@@ -39,6 +43,11 @@ class ResponseStreamDataType(str, Enum):
     def __str__(self) -> str:
         return self.value
 
+class BuiltinToolset(str, Enum):
+    NotebookEdit = 'nbi-notebook-edit'
+    NotebookExecute = 'nbi-notebook-execute'
+    PythonFileEdit = 'nbi-python-file-edit'
+
 class Signal:
     def __init__(self):
         self._listeners = []
@@ -70,9 +79,17 @@ class CancelToken:
     def cancellation_signal(self) -> Signal:
         return self._cancellation_signal
 
+@dataclass
+class RequestToolSelection:
+    built_in_toolsets: list[str] = None
+    mcp_server_tools: dict[str, list[str]] = None
+    extension_tools: dict[str, dict[str, list[str]]] = None
+
 @dataclass
 class ChatRequest:
     host: 'Host' = None
+    chat_mode: 'ChatMode' = None
+    tool_selection: RequestToolSelection = None
     command: str = ''
     prompt: str = ''
     chat_history: list[dict] = None
@@ -100,6 +117,14 @@ class MarkdownPartData(ResponseStreamData):
     def data_type(self) -> ResponseStreamDataType:
         return ResponseStreamDataType.MarkdownPart
 
+@dataclass
+class ImageData(ResponseStreamData):
+    content: str = ''
+
+    @property
+    def data_type(self) -> ResponseStreamDataType:
+        return ResponseStreamDataType.Image
+
 @dataclass
 class HTMLFrameData(ResponseStreamData):
     source: str = ''
@@ -281,9 +306,123 @@ class Tool:
     def pre_invoke(self, request: ChatRequest, tool_args: dict) -> Union[ToolPreInvokeResponse, None]:
         return None
 
-    async def handle_tool_call(self, request: ChatRequest, response: ChatResponse, tool_context: dict, tool_args: dict) -> dict:
+    async def handle_tool_call(self, request: ChatRequest, response: ChatResponse, tool_context: dict, tool_args: dict) -> str:
         raise NotImplemented
 
+class Toolset:
+    def __init__(self, id: str, name: str, description: str, provider: Union['NotebookIntelligenceExtension', None], tools: list[Tool] = [], instructions: str = None):
+        self.id = id
+        self.name = name
+        self.description = description
+        self.provider = provider
+        self.tools: list[Tool] = tools
+        self.instructions: Union[str, None] = instructions
+
+    def add_tool(self, tool: Tool) -> None:
+        self.tools.append(tool)
+
+    def remove_tool(self, tool: Tool) -> None:
+        self.tools.remove(tool)
+
+class SimpleTool(Tool):
+    def __init__(self, tool_function: Callable, name: str, description: str, schema: dict, title: str = None, auto_approve: bool = False, has_var_args: bool = False):
+        super().__init__()
+        self._tool_function = tool_function
+        self._name = name
+        self._description = description
+        self._schema = schema
+        self._title = title
+        self._auto_approve = auto_approve
+        self._has_var_args = has_var_args
+
+    @property
+    def name(self) -> str:
+        return self._name
+
+    @property
+    def title(self) -> str:
+        return self._title if self._title is not None else self._name
+    
+    @property
+    def tags(self) -> list[str]:
+        return []
+    
+    @property
+    def description(self) -> str:
+        return self._description
+    
+    @property
+    def schema(self) -> dict:
+        return self._schema
+
+    def pre_invoke(self, request: ChatRequest, tool_args: dict) -> Union[ToolPreInvokeResponse, None]:
+        confirmationTitle = None
+        confirmationMessage = None
+        if not self._auto_approve:
+            confirmationTitle = "Approve"
+            confirmationMessage = "Are you sure you want to call this tool?"
+        return ToolPreInvokeResponse(f"Calling tool '{self.name}'", confirmationTitle, confirmationMessage)
+
+    async def handle_tool_call(self, request: ChatRequest, response: ChatResponse, tool_context: dict, tool_args: dict) -> str:
+        fn_args = tool_args.copy()
+        if self._has_var_args:
+            fn_args.update({"request": request, "response": response})
+        return await self._tool_function(**fn_args)
+
+class MCPServer:
+    @property
+    def name(self) -> str:
+        return NotImplemented
+    
+    async def connect(self):
+        return NotImplemented
+
+    async def disconnect(self):
+        return NotImplemented
+    
+    def get_tools(self) -> list[Tool]:
+        return NotImplemented
+
+    def get_tool(self, tool_name: str) -> Tool:
+        return NotImplemented
+
+    async def call_tool(self, tool_name: str, tool_args: dict):
+        return NotImplemented
+
+def auto_approve(tool: SimpleTool):
+    """
+    Decorator to set auto_approve to True for a tool.
+    """
+    tool._auto_approve = True
+    return tool
+
+def tool(tool_function: Callable) -> SimpleTool:
+    mcp_tool = MCPToolClass.from_function(tool_function)
+    has_var_args = False
+    if "args" in mcp_tool.parameters["properties"]:
+        has_var_args = True
+        del mcp_tool.parameters["properties"]["args"]
+    if "args" in mcp_tool.parameters["required"]:
+        mcp_tool.parameters["required"].remove("args")
+
+    schema = {
+        "type": "function",
+        "function": {
+            "name": mcp_tool.name,
+            "description": mcp_tool.description,
+            "strict": False,
+            "parameters": mcp_tool.parameters
+        },
+    }
+
+    return SimpleTool(tool_function, mcp_tool.name, mcp_tool.description, schema, mcp_tool.name, auto_approve, has_var_args)
+
+class ChatMode:
+    def __init__(self, id: str, name: str, instructions: str = None):
+        self.id = id
+        self.name = name
+        self.instructions = instructions
+
 class ChatParticipant:
     @property
     def id(self) -> str:
@@ -322,6 +461,12 @@ class ChatParticipant:
 
         messages = request.chat_history.copy()
 
+        system_prompt = options.get("system_prompt")
+        if system_prompt is not None:
+            messages = [
+                {"role": "system", "content": system_prompt}
+            ] + messages
+
         if len(tools) == 0:
             request.host.chat_model.completions(messages, tools=None, cancel_token=request.cancel_token, response=response)
             return
@@ -335,22 +480,29 @@ class ChatParticipant:
 
         async def _tool_call_loop(tool_call_rounds: list):
             try:
+                if request.cancel_token.is_cancel_requested:
+                    return
+
                 tool_response = request.host.chat_model.completions(messages, openai_tools, cancel_token=request.cancel_token, options=options)
                 # after first call, set tool_choice to auto
                 options['tool_choice'] = 'auto'
 
-                if tool_response['choices'][0]['message'].get('tool_calls', None) is not None:
-                    for tool_call in tool_response['choices'][0]['message']['tool_calls']:
-                        tool_call_rounds.append(tool_call)
-                elif tool_response['choices'][0]['message'].get('content', None) is not None:
-                    response.stream(MarkdownData(tool_response['choices'][0]['message']['content']))
+                for choice in tool_response['choices']:
+                    if choice['message'].get('tool_calls', None) is not None:
+                        for tool_call in choice['message']['tool_calls']:
+                            tool_call_rounds.append(tool_call)
+                    elif choice['message'].get('content', None) is not None:
+                        response.stream(MarkdownData(tool_response['choices'][0]['message']['content']))
 
-                messages.append(tool_response['choices'][0]['message'])
+                    messages.append(choice['message'])
 
                 had_tool_call = len(tool_call_rounds) > 0
 
                 # handle first tool calls
                 while len(tool_call_rounds) > 0:
+                    if request.cancel_token.is_cancel_requested:
+                        return
+
                     tool_call = tool_call_rounds[0]
                     if "id" not in tool_call:
                         tool_call['id'] = uuid.uuid4().hex
@@ -379,11 +531,6 @@ class ChatParticipant:
                         else:
                             args = {}
 
-                    if len(tool_properties) != len(args):
-                        response.stream(MarkdownData(f"Oops! There was a problem handling tool request. Please try again with a different prompt."))
-                        response.finish()
-                        return
-
                     tool_pre_invoke_response = tool_to_call.pre_invoke(request, args)
                     if tool_pre_invoke_response is not None:
                         if tool_pre_invoke_response.message is not None:
@@ -404,7 +551,7 @@ class ChatParticipant:
 
                     function_call_result_message = {
                         "role": "tool",
-                        "content": json.dumps(tool_call_response),
+                        "content": str(tool_call_response),
                         "tool_call_id": tool_call['id']
                     }
 
@@ -486,7 +633,7 @@ class AIModel(LLMPropertyProvider):
         raise NotImplemented
 
     @property
-    def provider(self) -> str:
+    def provider(self) -> 'LLMProvider':
         return self._provider
     
     @property
@@ -551,6 +698,36 @@ class LLMProvider(LLMPropertyProvider):
                 return model
         return None
 
+class TelemetryEventType(str, Enum):
+    InlineCompletionRequest = 'inline-completion-request'
+    ExplainThisRequest = 'explain-this-request'
+    FixThisCodeRequest = 'fix-this-code-request'
+    ExplainThisOutputRequest = 'explain-this-output-request'
+    TroubleshootThisOutputRequest = 'troubleshoot-this-output-request'
+    GenerateCodeRequest = 'generate-code-request'
+    ChatRequest = 'chat-request'
+    InlineChatRequest = 'inline-chat-request'
+    ChatResponse = 'chat-response'
+    InlineChatResponse = 'inline-chat-response'
+    InlineCompletionResponse = 'inline-completion-response'
+
+class TelemetryEvent:
+    @property
+    def type(self) -> TelemetryEventType:
+        raise NotImplemented
+    
+    @property
+    def data(self) -> dict:
+        return None
+
+class TelemetryListener:
+    @property
+    def name(self) -> str:
+        raise NotImplemented
+
+    def on_telemetry_event(self, event: TelemetryEvent):
+        raise NotImplemented
+
 class Host:
     def register_llm_provider(self, provider: LLMProvider) -> None:
         raise NotImplemented
@@ -561,6 +738,16 @@ class Host:
     def register_completion_context_provider(self, provider: CompletionContextProvider) -> None:
         raise NotImplemented
     
+    def register_telemetry_listener(self, listener: TelemetryListener) -> None:
+        raise NotImplemented
+
+    def register_toolset(self, toolset: Toolset) -> None:
+        raise NotImplemented
+
+    @property
+    def nbi_config(self) -> NBIConfig:
+        raise NotImplemented
+
     @property
     def default_chat_participant(self) -> ChatParticipant:
         raise NotImplemented
@@ -577,6 +764,18 @@ class Host:
     def embedding_model(self) -> EmbeddingModel:
         raise NotImplemented
 
+    def get_mcp_server(self, server_name: str) -> MCPServer:
+        return NotImplemented
+
+    def get_mcp_server_tool(self, server_name: str, tool_name: str) -> Tool:
+        return NotImplemented
+
+    def get_extension_toolset(self, extension_id: str, toolset_id: str) -> Toolset:
+        return NotImplemented
+
+    def get_extension_tool(self, extension_id: str, toolset_id: str, tool_name: str) -> Tool:
+        return NotImplemented
+
 class NotebookIntelligenceExtension:
     @property
     def id(self) -> str:
