@@ -4,12 +4,14 @@ import asyncio
 from dataclasses import dataclass
 import json
 from os import path
+import datetime as dt
 import os
 from typing import Union
 import uuid
 import threading
 import logging
 import tiktoken
+from mcp.client.sse import sse_client
 
 from jupyter_server.extension.application import ExtensionApp
 from jupyter_server.base.handlers import APIHandler
@@ -17,21 +19,57 @@ from jupyter_server.utils import url_path_join
 import tornado
 from tornado import websocket
 from traitlets import Unicode
-from notebook_intelligence.api import CancelToken, ChatResponse, ChatRequest, ContextRequest, ContextRequestType, RequestDataType, ResponseStreamData, ResponseStreamDataType, BackendMessageType, Signal, SignalImpl
+from notebook_intelligence.api import BuiltinToolset, CancelToken, ChatMode, ChatResponse, ChatRequest, ContextRequest, ContextRequestType, RequestDataType, RequestToolSelection, ResponseStreamData, ResponseStreamDataType, BackendMessageType, SignalImpl
 from notebook_intelligence.ai_service_manager import AIServiceManager
-from notebook_intelligence.config import NBIConfig
 import notebook_intelligence.github_copilot as github_copilot
+from notebook_intelligence.built_in_toolsets import built_in_toolsets
 
 ai_service_manager: AIServiceManager = None
 log = logging.getLogger(__name__)
 tiktoken_encoding = tiktoken.encoding_for_model('gpt-4o')
 
 class GetCapabilitiesHandler(APIHandler):
+    notebook_execute_tool = 'enabled'
+
     @tornado.web.authenticated
     def get(self):
         ai_service_manager.update_models_from_config()
         nbi_config = ai_service_manager.nbi_config
         llm_providers = ai_service_manager.llm_providers.values()
+        notebook_execute_tool_enabled = self.notebook_execute_tool == 'enabled' or (self.notebook_execute_tool == 'env_enabled' and os.getenv('NBI_NOTEBOOK_EXECUTE_TOOL', 'disabled') == 'enabled')
+        allowed_builtin_toolsets = [{"id": toolset.id, "name": toolset.name} for toolset in built_in_toolsets.values() if toolset.id != BuiltinToolset.NotebookExecute or notebook_execute_tool_enabled]
+        mcp_servers = ai_service_manager.get_mcp_servers()
+        mcp_server_tools = [{"id": mcp_server.name, "tools": [{"name": tool.name, "description": tool.description} for tool in mcp_server.get_tools()]} for mcp_server in mcp_servers]
+        mcp_server_tools = [tool for tool in mcp_server_tools if len(tool["tools"]) > 0]
+        # sort by server id
+        mcp_server_tools.sort(key=lambda server: server["id"])
+
+        extensions = []
+        for extension_id, toolsets in ai_service_manager.get_extension_toolsets().items():
+            ts = []
+            for toolset in toolsets:
+                tools = []
+                for tool in toolset.tools:
+                    tools.append({"name": tool.name, "description": tool.description})
+                # sort by tool name
+                tools.sort(key=lambda tool: tool["name"])
+                ts.append({
+                    "id": toolset.id,
+                    "name": toolset.name,
+                    "description": toolset.description,
+                    "tools": tools
+                })
+            # sort by toolset name
+            ts.sort(key=lambda toolset: toolset["name"])
+            extension = ai_service_manager.get_extension(extension_id)
+            extensions.append({
+                "id": extension_id,
+                "name": extension.name,
+                "toolsets": ts
+            })
+        # sort by extension id
+        extensions.sort(key=lambda extension: extension["id"])
+
         response = {
             "using_github_copilot_service": nbi_config.using_github_copilot_service,
             "llm_providers": [{"id": provider.id, "name": provider.name} for provider in llm_providers],
@@ -40,8 +78,14 @@ class GetCapabilitiesHandler(APIHandler):
             "embedding_models": ai_service_manager.embedding_model_ids,
             "chat_model": nbi_config.chat_model,
             "inline_completion_model": nbi_config.inline_completion_model,
-            "embedding_model": nbi_config.embedding_model_id,
-            "chat_participants": []
+            "embedding_model": nbi_config.embedding_model,
+            "chat_participants": [],
+            "store_github_access_token": nbi_config.store_github_access_token,
+            "tool_config": {
+                "builtinToolsets": allowed_builtin_toolsets,
+                "mcpServers": mcp_server_tools,
+                "extensions": extensions
+            }
         }
         for participant_id in ai_service_manager.chat_participants:
             participant = ai_service_manager.chat_participants[participant_id]
@@ -58,10 +102,15 @@ class ConfigHandler(APIHandler):
     @tornado.web.authenticated
     def post(self):
         data = json.loads(self.request.body)
-        valid_keys = set(["chat_model", "inline_completion_model"])
+        valid_keys = set(["chat_model", "inline_completion_model", "store_github_access_token"])
         for key in data:
             if key in valid_keys:
                 ai_service_manager.nbi_config.set(key, data[key])
+                if key == "store_github_access_token":
+                    if data[key]:
+                        github_copilot.store_github_access_token()
+                    else:
+                        github_copilot.delete_stored_github_access_token()
         ai_service_manager.nbi_config.save()
         ai_service_manager.update_models_from_config()
         self.finish(json.dumps({}))
@@ -74,6 +123,14 @@ class UpdateProviderModelsHandler(APIHandler):
             ai_service_manager.ollama_llm_provider.update_chat_model_list()
         self.finish(json.dumps({}))
 
+class EmitTelemetryEventHandler(APIHandler):
+    @tornado.web.authenticated
+    def post(self):
+        event = json.loads(self.request.body)
+        thread = threading.Thread(target=asyncio.run, args=(ai_service_manager.emit_telemetry_event(event),))
+        thread.start()
+        self.finish(json.dumps({}))
+
 class GetGitHubLoginStatusHandler(APIHandler):
     # The following decorator should be present on all verb methods (head, get, post,
     # patch, put, delete, options) to ensure only authorized user can request the
@@ -177,6 +234,21 @@ class WebsocketCopilotResponseEmitter(ChatResponse):
                     }
                 ]
             }
+        elif data_type == ResponseStreamDataType.Image:
+            data = {
+                "choices": [
+                    {
+                        "delta": {
+                            "nbiContent": {
+                                "type": data_type,
+                                "content": data.content
+                            },
+                            "content": "",
+                            "role": "assistant"
+                        }
+                    }
+                ]
+            }
         elif data_type == ResponseStreamDataType.HTMLFrame:
             data = {
                 "choices": [
@@ -259,7 +331,7 @@ class WebsocketCopilotResponseEmitter(ChatResponse):
                                     "message": data.message,
                                     "confirmArgs": data.confirmArgs if data.confirmArgs is not None else {},
                                     "cancelArgs": data.cancelArgs if data.cancelArgs is not None else {},
-                                    "confirmLabel": data.confirmLabel if data.confirmLabel is not None else "Proceed",
+                                    "confirmLabel": data.confirmLabel if data.confirmLabel is not None else "Approve",
                                     "cancelLabel": data.cancelLabel if data.cancelLabel is not None else "Cancel"
                                 }
                             },
@@ -269,6 +341,25 @@ class WebsocketCopilotResponseEmitter(ChatResponse):
                     }
                 ]
             }
+        elif data_type == ResponseStreamDataType.MarkdownPart:
+            content = data.content
+            data = {
+                "choices": [
+                    {
+                        "delta": {
+                            "nbiContent": {
+                                "type": data_type,
+                                "content": data.content
+                            },
+                            "content": "",
+                            "role": "assistant"
+                        }
+                    }
+                ]
+            }
+            part = content
+            if part is not None:
+                self.streamed_contents.append(part)
         else: # ResponseStreamDataType.LLMRaw
             if len(data.get("choices", [])) > 0:
                 part = data["choices"][0].get("delta", {}).get("content", "")
@@ -279,7 +370,8 @@ class WebsocketCopilotResponseEmitter(ChatResponse):
             "id": self.messageId,
             "participant": self.participant_id,
             "type": BackendMessageType.StreamMessage,
-            "data": data
+            "data": data,
+            "created": dt.datetime.now().isoformat()
         })
 
     def finish(self) -> None:
@@ -343,6 +435,13 @@ class WebsocketCopilotHandler(websocket.WebSocketHandler):
             language = data['language']
             filename = data['filename']
             additionalContext = data.get('additionalContext', [])
+            chat_mode = ChatMode('agent', 'Agent') if data.get('chatMode', 'ask') == 'agent' else ChatMode('ask', 'Ask')
+            toolSelections = data.get('toolSelections', {})
+            tool_selection = RequestToolSelection(
+                built_in_toolsets=toolSelections.get('builtinToolsets', []),
+                mcp_server_tools=toolSelections.get('mcpServers', {}),
+                extension_tools=toolSelections.get('extensions', {})
+            )
 
             request_chat_history = self.chat_history.get_history(chatId).copy()
 
@@ -372,7 +471,7 @@ class WebsocketCopilotHandler(websocket.WebSocketHandler):
             response_emitter = WebsocketCopilotResponseEmitter(chatId, messageId, self, self.chat_history)
             cancel_token = CancelTokenImpl()
             self._messageCallbackHandlers[messageId] = MessageCallbackHandlers(response_emitter, cancel_token)
-            thread = threading.Thread(target=asyncio.run, args=(ai_service_manager.handle_chat_request(ChatRequest(prompt=prompt, chat_history=request_chat_history, cancel_token=cancel_token), response_emitter),))
+            thread = threading.Thread(target=asyncio.run, args=(ai_service_manager.handle_chat_request(ChatRequest(chat_mode=chat_mode, tool_selection=tool_selection, prompt=prompt, chat_history=request_chat_history, cancel_token=cancel_token), response_emitter),))
             thread.start()
         elif messageType == RequestDataType.GenerateCode:
             data = msg['data']
@@ -383,6 +482,7 @@ class WebsocketCopilotHandler(websocket.WebSocketHandler):
             existing_code = data['existingCode']
             language = data['language']
             filename = data['filename']
+            chat_mode = ChatMode('ask', 'Ask')
             if prefix != '':
                 self.chat_history.add_message(chatId, {"role": "user", "content": f"This code section comes before the code section you will generate, use as context. Leading content: ```{prefix}```"})
             if suffix != '':
@@ -394,7 +494,7 @@ class WebsocketCopilotHandler(websocket.WebSocketHandler):
             cancel_token = CancelTokenImpl()
             self._messageCallbackHandlers[messageId] = MessageCallbackHandlers(response_emitter, cancel_token)
             existing_code_message = " Update the existing code section and return a modified version. Don't just return the update, recreate the existing code section with the update." if existing_code != '' else ''
-            thread = threading.Thread(target=asyncio.run, args=(ai_service_manager.handle_chat_request(ChatRequest(prompt=prompt, chat_history=self.chat_history.get_history(chatId), cancel_token=cancel_token), response_emitter, options={"system_prompt": f"You are an assistant that generates code for '{language}' language. You generate code between existing leading and trailing code sections.{existing_code_message} Be concise and return only code as a response. Don't include leading content or trailing content in your response, they are provided only for context. You can reuse methods and symbols defined in leading and trailing content."}),))
+            thread = threading.Thread(target=asyncio.run, args=(ai_service_manager.handle_chat_request(ChatRequest(chat_mode=chat_mode, prompt=prompt, chat_history=self.chat_history.get_history(chatId), cancel_token=cancel_token), response_emitter, options={"system_prompt": f"You are an assistant that generates code for '{language}' language. You generate code between existing leading and trailing code sections.{existing_code_message} Be concise and return only code as a response. Don't include leading content or trailing content in your response, they are provided only for context. You can reuse methods and symbols defined in leading and trailing content."}),))
             thread.start()
         elif messageType == RequestDataType.InlineCompletionRequest:
             data = msg['data']
@@ -417,7 +517,7 @@ class WebsocketCopilotHandler(websocket.WebSocketHandler):
                 return
             handlers.response_emitter.on_user_input(msg['data'])
         elif messageType == RequestDataType.ClearChatHistory:
-            self.chat_history.clear(msg['data']['chatId'])
+            self.chat_history.clear()
         elif messageType == RequestDataType.RunUICommandResponse:
             handlers = self._messageCallbackHandlers.get(messageId)
             if handlers is None:
@@ -463,15 +563,14 @@ class NotebookIntelligence(ExtensionApp):
     handlers = []
     root_dir = ''
 
-    github_access_token = Unicode(
-        default_value=None,
+    notebook_execute_tool = Unicode(
+        default_value="enabled",
         help="""
-        GitHub Copilot credentials storage options.
-
-        'remember' - Store credentials in keyring and remember next time.
-        'forget' - Remove credentials stored in keyring.
-        '<TOKEN_VALUE>' - Use the provided access token value.
+        Notebook execute tool options.
 
+        'enabled' - Enable notebook execute tool (default).
+        'disabled' - Disabled notebook execute tool.
+        'env_enabled' - Disabled by default, can be enabled using 'NBI_NOTEBOOK_EXECUTE_TOOL=enabled'.
         """,
         allow_none=True,
         config=True,
@@ -482,13 +581,14 @@ class NotebookIntelligence(ExtensionApp):
 
     def initialize_handlers(self):
         NotebookIntelligence.root_dir = self.serverapp.root_dir
-        self.initialize_ai_service()
+        server_root_dir = os.path.expanduser(self.serverapp.web_app.settings["server_root_dir"])
+        self.initialize_ai_service(server_root_dir)
         self._setup_handlers(self.serverapp.web_app)
         self.serverapp.log.info(f"Registered {self.name} server extension")
     
-    def initialize_ai_service(self):
+    def initialize_ai_service(self, server_root_dir: str):
         global ai_service_manager
-        ai_service_manager = AIServiceManager({"github_access_token": self.github_access_token})
+        ai_service_manager = AIServiceManager({"server_root_dir": server_root_dir})
 
     def initialize_templates(self):
         pass
@@ -504,14 +604,17 @@ class NotebookIntelligence(ExtensionApp):
         route_pattern_capabilities = url_path_join(base_url, "notebook-intelligence", "capabilities")
         route_pattern_config = url_path_join(base_url, "notebook-intelligence", "config")
         route_pattern_update_provider_models = url_path_join(base_url, "notebook-intelligence", "update-provider-models")
+        route_pattern_emit_telemetry_event = url_path_join(base_url, "notebook-intelligence", "emit-telemetry-event")
         route_pattern_github_login_status = url_path_join(base_url, "notebook-intelligence", "gh-login-status")
         route_pattern_github_login = url_path_join(base_url, "notebook-intelligence", "gh-login")
         route_pattern_github_logout = url_path_join(base_url, "notebook-intelligence", "gh-logout")
         route_pattern_copilot = url_path_join(base_url, "notebook-intelligence", "copilot")
+        GetCapabilitiesHandler.notebook_execute_tool = self.notebook_execute_tool
         NotebookIntelligence.handlers = [
             (route_pattern_capabilities, GetCapabilitiesHandler),
             (route_pattern_config, ConfigHandler),
             (route_pattern_update_provider_models, UpdateProviderModelsHandler),
+            (route_pattern_emit_telemetry_event, EmitTelemetryEventHandler),
             (route_pattern_github_login_status, GetGitHubLoginStatusHandler),
             (route_pattern_github_login, PostGitHubLoginHandler),
             (route_pattern_github_logout, GetGitHubLogoutHandler),
